import numpy as np
from scipy.optimize import OptimizeResult, linprog

from qitip.objects import EntropicSpace

# Prover is created with Quantum Elemental Inequalities
# In principle, it can also be created with classical elemental inequalities
from qitip.quantum_inequalities import QuantumElementalInequalities


class Prover:
    def __init__(self, space: EntropicSpace):
        self.n: int = max(max(space.vector_entry))
        self.elemental = QuantumElementalInequalities(
            space.vector_entry
        ).get_elementals()

        # This is only used for method isVonNeumannType
        self._vector_entry = space.vector_entry

    def __hash__(self) -> int:
        return hash((self.n))

    def _check_type(
        self,
        inequality: np.ndarray[
            np.ndarray[float, np.dtype[np.float64]], np.dtype[np.float64]
        ],
        constraints: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
    ) -> bool:
        """_summary_
        In classical ITIP, checking if an inequality is Shannon-type can be proved with linear programming.

        In quantum information, a similar approach is applied.

        In summary, if the maximum value of the dual problem is 0, the inequality is von-Neumann type;
        otherwise, it is not provable by quantum ITIP.

        Args:
            inequality (np.ndarray[ np.ndarray[float, np.dtype[np.float64]], np.dtype[np.float64] ]):
            constraints (np.ndarray[ np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64] ]):

        Raises:
            ValueError: when scipy successfully get the optimal value which fails to be 0. Inconsistent with the
            theory

        Returns:
            bool: if the inequality under the constraints is von-Neumann type, it returns True; otherwise, it returns False.
        """
        max_value: int = 0

        result: OptimizeResult = linprog(
            c=-np.zeros(self.elemental.shape[0] + constraints.shape[0]),
            A_eq=np.vstack((self.elemental, -constraints)).transpose(),
            b_eq=inequality,
            bounds=tuple(
                [(0, None)] * self.elemental.shape[0]
                + [(None, None)] * constraints.shape[0]
            ),
        )
        if not result.success or (result.success and (result.fun == max_value)):
            return result.success
        else:
            raise ValueError(
                f"Unexpected error has occurred. Expected optimal value is {max_value}, but get {max.func} instead."
            )

    def _shortest_proof(
        self,
        inequality: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
        constraints: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
    ) -> tuple[
        np.ndarray[np.float64, np.dtype[np.float64]],
        np.ndarray[np.float64, np.dtype[np.float64]],
    ]:
        # The vectors are in the order of [y, mu, t] where
        # |y|_{0} = num of elemental inequalities
        # |mu|_{0} = |t|_{0} = num of constraints
        # Also -t <= mu <= t is the constraint

        num_elementals: int = self.elemental.shape[0]
        num_constraints: int = constraints.shape[0]
        dim: int = len(self._vector_entry)

        # Inequality constraints: -t <= mu <= t
        # Scipy: [A_ub][x] <= [b_ub]
        ## First, -t <= mu
        l_ineq = np.concatenate(
            (
                np.zeros((num_constraints, num_elementals)),
                -np.identity(num_constraints),
                -np.identity(num_constraints),
            ),
            axis=1,
        )
        ## Next, mu <= t
        r_ineq: np.ndarray[
            np.ndarray[np.int64, np.dtype[np.int64]], np.dtype[np.int64]
        ] = np.concatenate(
            (
                np.zeros((num_constraints, num_elementals)),
                np.identity(num_constraints),
                -np.identity(num_constraints),
            ),
            axis=1,
        )

        result = linprog(
            c=np.array(
                [1] * num_elementals + [0] * num_constraints + [1] * num_constraints
            ),
            A_ub=np.concatenate((l_ineq, r_ineq), axis=0),
            b_ub=np.zeros(2 * num_constraints),
            A_eq=np.concatenate(
                (
                    self.elemental,
                    -constraints,
                    np.zeros((num_constraints, dim)),
                ),
                axis=0,
            ).transpose(),
            b_eq=inequality,
            bounds=tuple(
                [(0, None)] * num_elementals
                + [(None, None)] * num_constraints
                + [(0, None)] * num_constraints
            ),
        )

        if not result.success:
            raise ValueError("Solution to shortest proof is not found ... ")

        return (
            result.x[:num_elementals],
            result.x[num_elementals : num_elementals + num_constraints],
        )

    # In theory, the maximal value of the dual problem cannot be found
    # We have to restrict our search in a bounded region.
    # Different from the classical information theory, marginal entropies are bounded by
    # H(all random variable)
    def _counter_proof_gamma(
        self,
        inequality: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
        constraints: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
    ) -> np.ndarray[np.float64, np.dtype[np.float64]]:
        # In quantum information theory, the closest counterpart is that a S(I) <= sum_{i in I}S({i})
        # Hence, I require, S({i}) <= 1 where i in {1,2,3,...,n}

        # The vector we will be working with is in the order of [y, mu, gamma]
        result: OptimizeResult = linprog(
            c=np.array(
                [0] * (self.elemental.shape[0] + constraints.shape[0]) + [1] * self.n
            ),
            A_eq=np.concatenate(
                (
                    self.elemental,
                    -constraints,
                    -np.eye(N=self.n, M=self.elemental.shape[1]),
                ),
                axis=0,
            ).transpose(),
            b_eq=inequality,
            bounds=tuple(
                [(0, None)] * self.elemental.shape[0]
                + [(None, None)] * constraints.shape[0]
                + [(0, None)] * self.n
            ),
            method="interior-point",  # This yields more desiring result than the default,  HiGHS
        )

        if result.status is False:
            raise ValueError(
                "Expect optimal value of dual problem under the bounded marginal entropies to be found ..."
            )

        return result.x[-self.n :]

    def _shortest_counter_proof(
        self,
        inequality: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
        constraints: np.ndarray[
            np.ndarray[np.float64, np.dtype[np.float64]], np.dtype[np.float64]
        ],
    ) -> tuple[
        np.ndarray[np.float64, np.dtype[np.float64]],
        np.ndarray[np.float64, np.dtype[np.float64]],
    ]:
        # This is quite complicated, for more information, checkout my master's thesis about
        # the theory of quantum ITIP
        gamma: np.ndarray[np.float64, np.dtype[np.float64]] = self._counter_proof_gamma(
            inequality=inequality, constraints=constraints
        )

        return self._shortest_proof(
            inequality=inequality
            + np.matmul(gamma, np.eye(N=self.n, M=self.elemental.shape[1])),
            constraints=constraints,
        )


class ProverPool:
    def __init__(self) -> None:
        self._created: set[Prover] = set()

    def get(self, space: EntropicSpace):
        for p in self._created:
            if p.n == space.n:
                return p

        new_prover = Prover(space=space)
        self._created.add(new_prover)
        return new_prover



from typing import Iterable

from qitip.objects import (
    Constraints,
    ConstraintsBuilder,
    Inequality,
    InequalityBuilder,
)
from qitip.utils.converters import CoefficientsToDict

# Test with bipartite system for simplicity
vector_entry: dict[frozenset[int], int] = {
    frozenset({1}): 0,
    frozenset({2}): 1,
    frozenset({1, 2}): 2,
}


def test_convert_inequality_coefficients_to_dict() -> None:
    vec: dict[Iterable[int] | int, float] = {(1, 2): 1, (1,): -1}

    builder = InequalityBuilder(vector_entry)
    inq: Inequality = builder.from_coefficients(vec)

    assert vec == CoefficientsToDict.convert_vector(inq.vector_entry, inq.coefficients)


def test_convert_empty_constraints_to_dict() -> None:
    vec = [
        dict(),
    ]

    builder = ConstraintsBuilder(vector_entry)

    empty_constraint: Constraints = builder.from_coefficients(vec)

    # Note that this returns an empty list rather than vec
    assert [] == CoefficientsToDict.convert_matrix(
        empty_constraint.vector_entry, empty_constraint.coefficients
    )


def test_convert_single_constraint_to_dict() -> None:
    builder = ConstraintsBuilder(vector_entry)

    c: tuple[dict[Iterable[int] | int, float]] = ({(1, 2): 1, (1,): -1},)

    single_constraints: Constraints = builder.from_coefficients(c)

    assert list(c) == CoefficientsToDict.convert_matrix(
        single_constraints.vector_entry, single_constraints.coefficients
    )


def test_convert_multiple_constraints_to_dict() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    c: tuple[dict[Iterable[int] | int, float], ...] = (
        {(1, 2): 1, (1,): -1},
        {(1,): -1, (2,): 1},
    )

    multi_constraints = builder.from_coefficients(c)

    assert list(c) == CoefficientsToDict.convert_matrix(
        multi_constraints.vector_entry, multi_constraints.coefficients
    )



import numpy as np
import pytest
from qitip.objects import EntropicSpace
from qitip.quantum_inequalities import QuantumElementalInequalities


def test_one_system_all_inequalities():
    n: int = 1

    space: EntropicSpace = EntropicSpace(n=n)

    with pytest.raises(ValueError):
        QuantumElementalInequalities(vector_entry=space.vector_entry).get_elementals()


def test_two_system_type_1() -> None:
    n: int = 2
    space = EntropicSpace(n=n)

    quantum_ineqality = QuantumElementalInequalities(vector_entry=space.vector_entry)
    intended_inequalities = np.array([[1, 1, -1]])

    type_1_inequalities = quantum_ineqality._get_all_type_1()
    num_inequalities = type_1_inequalities.shape[0]

    assert num_inequalities == 1
    assert all(
        [inequality in type_1_inequalities for inequality in intended_inequalities]
    )


def test_three_system_type_1() -> None:
    n: int = 3

    space = EntropicSpace(n)

    quantum_inequality = QuantumElementalInequalities(space.vector_entry)
    intended_inequalities = np.array(
        [
            [0, 0, -1, 0, 1, 1, -1],
            [1, 1, 0, -1, 0, 0, 0],
            [0, -1, 0, 1, 0, 1, -1],
            [1, 0, 1, 0, -1, 0, 0],
            [-1, 0, 0, 1, 1, 0, -1],
            [0, 1, 1, 0, 0, -1, 0],
        ]
    )

    type_1_inequalities = quantum_inequality._get_all_type_1()
    num_inequalities: int = type_1_inequalities.shape[0]

    assert num_inequalities == 6
    assert all(
        [inequality in type_1_inequalities for inequality in intended_inequalities]
    )


def test_ten_system_type_1() -> None:
    n: int = 10

    space = EntropicSpace(n)

    quantum_inequality = QuantumElementalInequalities(space.vector_entry)
    num_inequalities = quantum_inequality._get_all_type_1().shape[0]

    assert num_inequalities == 10 * 9 * 2 ** (7)


def test_two_system_type2() -> None:
    n: int = 2

    space = EntropicSpace(n=n)

    quantum_inequality = QuantumElementalInequalities(space.vector_entry)
    intended_inequalities = np.array([[1, -1, 1], [-1, 1, 1]])

    type_2_inequalities = quantum_inequality._get_all_type_2()
    num_inequalities_t2 = type_2_inequalities.shape[0]

    assert num_inequalities_t2 == 2
    assert all(
        [inequality in type_2_inequalities for inequality in intended_inequalities]
    )


def test_three_system_type2() -> None:
    n: int = 3

    space = EntropicSpace(n)

    quantum_inequality = QuantumElementalInequalities(space.vector_entry)
    intended_inequalities = np.array(
        [
            [0, -1, -1, 1, 0, 1, 0],
            [1, 0, 0, 0, 0, -1, 1],
            [-1, 0, -1, 1, 0, 1, 0],
            [0, 1, 0, 0, -1, 0, 1],
            [-1, -1, 0, 0, 1, 1, 0],
            [0, 0, 1, -1, 0, 0, 1],
        ]
    )

    type_2_inequalities = quantum_inequality._get_all_type_2()
    num_inequality_t2 = type_2_inequalities.shape[0]

    assert num_inequality_t2 == n * 2 ** (n - 2)
    assert all(
        [inequality in type_2_inequalities for inequality in intended_inequalities]
    )


def test_ten_system_type2() -> None:
    n: int = 10

    space = EntropicSpace(n)

    quantum_inequality = QuantumElementalInequalities(space.vector_entry)
    num_inequality_t2 = quantum_inequality._get_all_type_2().shape[0]

    assert num_inequality_t2 == n * 2 ** (n - 2)



import numpy as np
import pytest
from qitip.objects import Inequality

vector_entry: dict[frozenset[int], int] = {
    frozenset((1,)): 0,
    frozenset((2,)): 1,
    frozenset((1, 2)): 2,
}


def test_valid_creation():
    Inequality(vector_entry=vector_entry, v=(1, 0, 0))


def test_intended_numpy_array():
    vector = (1, 0, 0)
    inequality = Inequality(vector_entry=vector_entry, v=vector)
    assert (inequality.coefficients == np.array([vector])).all()
    assert (inequality.coefficients.shape) == (len(vector),)


def test_v_longer_than_entry_mapping():
    with pytest.raises(ValueError):
        Inequality(vector_entry=vector_entry, v=(1, 2, 3, 4))


def test_v_shorter_than_entry_mapping():
    with pytest.raises(ValueError):
        Inequality(vector_entry=vector_entry, v=(1, 2))



# Information inequality is studied in entropic space
# It is responsible for specifying the meaning for each axis
# Given an n-party quantum system, each axis represents an element in the superset of {1,2,...,n}
from dataclasses import dataclass, field
from itertools import combinations


@dataclass
class EntropicSpace:
    """
    Generate all possible pairs from n random variables.

    The class includes information of

    1. All the pairs

    2. The index in an entropic vector that corresponds to the given pairing
    """

    # Number of parties
    n: int
    _all_pairs: tuple[frozenset[int], ...] = field(init=False)
    _vector_entry: dict[frozenset[int], int] = field(init=False)

    def __hash__(self) -> int:
        return hash((self.n))

    def __post_init__(self):
        self._all_pairs: tuple[frozenset[int], ...] = self.generate_all_pairs(self.n)
        self._vector_entry: dict[frozenset[int], int] = {
            v: k for k, v in enumerate(self._all_pairs)
        }

    # Properties
    @property
    def vector_entry(self) -> dict[frozenset[int], int]:
        return self._vector_entry

    @staticmethod
    def generate_all_pairs(n: int) -> tuple[frozenset[int], ...]:
        labeling: set = set(range(1, n + 1))

        return tuple(
            frozenset(group)
            for size in range(1, n + 1)
            for group in combinations(labeling, r=size)
        )


class SpacePool:
    def __init__(self):
        self._created: set[EntropicSpace] = set()

    def get(self, n: int) -> EntropicSpace:
        for s in self._created:
            if s.n == n:
                return s

        new_space = EntropicSpace(n=n)
        self._created.add(new_space)
        return new_space



from typing import Iterable

import numpy as np
from qitip.objects import Constraints, Inequality
from qitip.qitip import Qitip


def test_embed_inequality_to_higher_dim() -> None:
    q2 = Qitip(n=2)
    q3 = Qitip(n=3)

    vec: dict[Iterable[int] | int, float] = {(1, 2): 1}

    inq_in_2: Inequality = q2.inequality.from_coefficients(vec)
    inq_in_3: Inequality = q3.embed(inq_in_2)

    expect_inq = q3.inequality.from_coefficients(vec)

    assert np.array_equiv(inq_in_3.coefficients, expect_inq.coefficients)


def test_embed_empty_constraints_to_higher_dim():
    q2 = Qitip(n=2)
    q3 = Qitip(n=3)

    empty_cons_in_2: Constraints = q2.constraints()
    empty_cnos_in_3: Constraints = q3.embed(empty_cons_in_2)

    assert np.array_equiv(empty_cnos_in_3.coefficients, q3.constraints().coefficients)


def test_embed_single_constraints_to_higher_dim() -> None:
    q2 = Qitip(n=2)
    q3 = Qitip(n=3)

    c: tuple[dict[Iterable[int] | int, float]] = ({(1, 2): 1, (1): -1},)

    one_constraint_in_2: Constraints = q2.constraints.from_coefficients(c)
    one_constraint_in_3: Constraints = q3.embed(one_constraint_in_2)

    assert np.array_equiv(
        one_constraint_in_3.coefficients,
        q3.constraints.from_coefficients(c).coefficients,
    )


def test_embed_multi_constraints_to_higher_dim() -> None:
    q2 = Qitip(n=2)
    q3 = Qitip(n=3)

    c: tuple[dict[Iterable[int] | int, float], ...] = (
        {(1, 2): 1, (1): -1},
        {(1,): -1, (2): 1},
    )

    multi_constraints_in_2: Constraints = q2.constraints.from_coefficients(c)
    multi_constraints_in_3: Constraints = q3.embed(multi_constraints_in_2)

    assert np.array_equiv(
        multi_constraints_in_3.coefficients,
        q3.constraints.from_coefficients(c).coefficients,
    )



# Inequality is responsible for storing the information of
# 1. the information inequality given by the user in the "canoncial form"
# 2. the mapping between entries in the vector and the marginal entropy

from dataclasses import InitVar, dataclass, field
from typing import Iterable, Optional

import numpy as np
from numpy.typing import ArrayLike

from qitip.utils.converters import create_vector_with_coefficient
from qitip.utils.validators import validate_vector


@dataclass
class Inequality:
    vector_entry: dict[frozenset[int], int]
    v: InitVar[ArrayLike]
    coefficients: np.ndarray = field(init=False)

    def __post_init__(self, v):
        self.coefficients = validate_vector(v=v, dim=len(self.vector_entry)).reshape(
            (-1,)
        )


class InequalityBuilder:
    # Inequality builder is chaacterized by the vector entry
    def __init__(self, vector_entry: dict[frozenset[int], int]):
        self._vector_entry: dict[frozenset[int], int] = vector_entry

    # When we use ib = InequalityBuilder(vectro_entry)
    # ib(v) should give us a new Inequality object
    def __call__(self, v: Optional[ArrayLike] = None) -> Inequality:
        if v is None:
            raise TypeError(
                f"Qitip(n).inequality(v) missing 1 required positional argument: 'v' which has the dimension of '{len(self._vector_entry)}'"
            )
        return Inequality(vector_entry=self._vector_entry, v=v)

    # As the number of coefficients increases, it is easier to
    # create an object by specifying the coefficients
    def from_coefficients(self, v: dict[Iterable[int] | int, float]) -> Inequality:
        return Inequality(
            vector_entry=self._vector_entry,
            v=create_vector_with_coefficient(vector_entry=self._vector_entry)(v),
        )



from qitip.main import init

while True:
    try:
        n: int = int(input("Input the number of quantum systems: "))
        if n >= 2:
            break
        else:
            print("Number of quantum systems has to be greater than 2 ...")
    except ValueError:
        print("Input value should be an integer greater than 1!")
    except Exception:
        raise Exception("Unexpected errors occur ...")

init(n)



from qitip.objects import (
    Constraints,
    ConstraintsBuilder,
    EntropicSpace,
    Inequality,
    InequalityBuilder,
    result_director,
)
from qitip.prover import Prover


def test_winter_linden_inequality() -> None:
    space = EntropicSpace(n=4)
    constraints: Constraints = ConstraintsBuilder(space.vector_entry).from_coefficients(
        [
            {(2, 3): 1, (1, 2): 1, (1, 2, 3): -1, (2): -1},
            {(1, 2): 1, (1, 3): 1, (1, 2, 3): -1, (1): -1},
            {(1, 4): 1, (2, 4): 1, (1, 2, 4): -1, (4): -1},
        ]
    )
    inequality: Inequality = InequalityBuilder(space.vector_entry).from_coefficients(
        {(4): 1, (3, 4): -1, (1, 2): -1, (1, 2, 3): 1}
    )

    prover: Prover = Prover(space)

    assert result_director(prover, inequality, constraints).status is False



from typing import Any

import numpy as np
from numpy.typing import ArrayLike


def validate_vector(
    v: ArrayLike, dim: int
) -> np.ndarray[np.float64, np.dtype[np.float64 | np.int64]]:
    vec: np.ndarray[Any, np.dtype[Any]] = np.array(v)

    vec_shape: tuple = vec.shape

    # check if the shape is that of a vector
    if len(vec_shape) > 1:
        raise ValueError(f"Incorrect shape, {vec_shape}, for a vector, (n, )")
    # check if the vector matches the dimensionality
    elif vec_shape[0] != dim:
        raise ValueError(
            f"Vector dimension, {vec_shape[0]}, inconsistent with the dimension of the system, {dim}"
        )
    # check if all the entries are scalars
    elif vec.dtype != np.float64 and vec.dtype != np.int64:
        raise TypeError("The entries can only be scalars.")
    else:
        return vec


def validate_matrix(
    m: ArrayLike, dim: int
) -> np.ndarray[np.float64, np.dtype[np.float64 | np.int64]]:
    matr: np.ndarray[Any, np.dtype[Any]] = np.array(m)
    matr_shape = matr.shape

    if len(matr_shape) > 2:
        raise ValueError(
            f"Input data is at most converted to a 2D matrix. Data dimension: {len(matr_shape)}"
        )

    elif len(matr_shape) == 1:
        return validate_vector(v=m, dim=dim)

    elif matr.dtype != np.float64 and matr.dtype != np.int64:
        raise TypeError("The entries can only be scalars.")

    else:
        return matr
